import hashlib
import os

class UnsupportedTypeError(Exception):
    pass

def get_hash(data=None, prefix="", suffix="", **kwargs):
    # 버전 호환: 이전 버전에서 data 변수가 아닌 text 로 받음
    if ("text" in kwargs) and (data is None):
        data=kwargs["text"]
        
    if isinstance(data, dict):
        # 딕셔너리인 경우, 키를 정렬하여 재귀적으로 처리
        items = sorted((k, get_hash(v)) for k, v in data.items())
        data_string = prefix + str(items) + suffix
    elif isinstance(data, list):
        # 리스트인 경우, 요소를 재귀적으로 처리
        items = [get_hash(item) for item in data]
        data_string = prefix + str(items) + suffix
    elif isinstance(data, (str, int, float, bool, type(None), bytes)):
        # 기본 타입인 경우, 문자열로 변환
        data_string = prefix + str(data) + suffix
    else:
        # 기타 자료형인 경우, 예외 발생
        raise UnsupportedTypeError(f"Unsupported type: {type(data).__name__}")

    # 해시값 생성
    encoded_string = data_string.encode('utf-8', errors='ignore')
    hash_value = hashlib.sha256(encoded_string).hexdigest()
    return hash_value


def get_hash_file(file_path, prefix="", suffix="", use_chunked=True, chunk_size=4096):
    sha256_hash = hashlib.sha256()
    
    # Prefix를 해시 업데이트
    sha256_hash.update(prefix.encode('utf-8', errors='ignore'))
    
    file_size = os.path.getsize(file_path)
    
    # 파일을 읽어와서 해시 업데이트
    with open(file_path, "rb") as f:
        if use_chunked and file_size > chunk_size:
            # 큰 파일을 처리할 때 부분적으로 읽어오는 방식
            for byte_block in iter(lambda: f.read(chunk_size), b""):
                sha256_hash.update(byte_block)
        else:
            # 작은 파일을 처리할 때 파일 전체를 한 번에 읽어오는 방식
            file_content = f.read()
            sha256_hash.update(file_content)
    
    # Suffix를 해시 업데이트
    sha256_hash.update(suffix.encode('utf-8', errors='ignore'))
    
    return sha256_hash.hexdigest()


def execute_python_code(code: str) -> dict:
    """
    사용자 시스템에서 실행 할 Python 코드를 작성하세요. (코드 내에 절대로 def 는 사용하면 안됩니다.)
    이 함수는 사용자 시스템에서 실행되며, 사용자의 파일 읽기, 수정 및 시스템을 제어할 수 있습니다.
    사용자에게 파일 업로드를 요청하지 말고 이 함수로 사용자 파일에 접근하여 code_interpreter 처럼 사용하세요.
    이전에 실행된 코드의 결과는 전역 세션 변수(global_session)에 저장되므로 지속적인 확인 및 참조가 가능합니다.
    사용자의 요청이 단계적 수행이 필요한 경우 단계적으로 function을 요청 하십시오.
    """

    global global_session
    if "global_session" not in globals():
        global_session = {
            "global_scope": {},
            "local_scope": {}
        }

    try:
        exec(code, global_session["global_scope"], global_session["local_scope"])
        result = {
            "success": True,
            "result": global_session["local_scope"].get("operation_result", None)
        }
    except Exception as e:
        result = {
            "success": False,
            "error": str(e)
        }
    
    return str(result)



import os
import requests
from PIL import Image
import base64
from io import BytesIO

class ImageLoader:
    def __init__(self, target, format=None):
        self.image = self.load_image(target)
        if not self.image:
            raise ValueError("이미지를 로드할 수 없습니다.")
        self.format = format if format else self.image.format
        self.base64_str = self.get_base64_str()
        self.base64_str_url = self.get_base64_str_url()
    
    def load_image(self, target):
        if isinstance(target, Image.Image) :
            return target
        elif isinstance(target, str):
            # 파일 경로에서 로드
            try :
                image = Image.open(target)
                return image
            except :
                pass
            
            # URL에서 로드
            try :
                response = requests.get(target)
                image_data = BytesIO(response.content)
                image = Image.open(image_data)
                return image
            except :
                pass
            
            # base64 문자열에서 로드
            try :
                image_data = BytesIO(base64.b64decode(target))
                image = Image.open(image_data)
                return image
            except :
                pass
        else :
            raise Exception("Invalid target")
    
    def get_base64_str(self):
        with BytesIO() as buffer:
            self.image.save(buffer, format=self.format)
            base64_str = base64.b64encode(buffer.getvalue()).decode('utf-8')
        return base64_str
    
    def get_base64_str_url(self):
        mime_type = f"image/{self.format.lower()}"
        return f"data:{mime_type};base64,{self.base64_str}"
        
    def convert(self, format=None, size=None, max_size=None):
        converted_format = format if format else self.format
        converted_image = self.image.copy()
        if converted_format != self.format  :
            if converted_format == 'JPEG':
                converted_image = converted_image.convert('RGB')
            elif converted_format == 'GIF':
                converted_image = converted_image.convert('P', palette=Image.ADAPTIVE)
            elif converted_format == 'PNG':
                converted_image = converted_image.convert('RGBA')
            else :
                raise Exception("Invalid format")
        if size :
            converted_image = converted_image.resize(size)
        elif max_size:
            width, height = converted_image.size
            if max(width, height) > max_size:
                if width > height:
                    new_width = max_size
                    new_height = int((max_size / width) * height)
                else:
                    new_height = max_size
                    new_width = int((max_size / height) * width)
                converted_image = converted_image.resize((new_width, new_height))
        return ImageLoader(converted_image, format=converted_format)

    def show(self):
        self.image.show()

    def save_image(self, save_path):
        self.image.save(save_path, format=self.format)


import os
from pathspec import PathSpec


def get_patterns(file_path):
    if os.path.exists(file_path):
        with open(file_path, 'r', encoding='utf-8') as f:
            patterns = f.readlines()
        patterns = [pattern.strip() for pattern in patterns if pattern.strip() and not pattern.startswith('#')]
    else:
        patterns = []
    return patterns

def get_all_filepaths(gitignore=True):
    if gitignore :
        patterns = get_patterns('./.gitignore')
    else :
        patterns = []

    spec = PathSpec.from_lines('gitwildmatch', patterns)
    all_filepaths = []
    for root, dirs, files in os.walk("./"):
        if '.git' in dirs:
            dirs.remove('.git')

        for name in files:
            filepath = os.path.relpath(os.path.join(root, name), start="./")
            all_filepaths.append(filepath)

    all_filepaths = [f for f in all_filepaths if not spec.match_file(f)]
    return sorted(all_filepaths)


def get_target_filepaths(patterns, ignore):
    all_filepaths = get_all_filepaths()
    patterns = [pattern.strip() for pattern in patterns if pattern.strip()]
    spec = PathSpec.from_lines('gitwildmatch', patterns)
    if ignore :
        return [f for f in all_filepaths if not spec.match_file(f)]
    else :
        return [f for f in all_filepaths if spec.match_file(f)]


def get_chat_filepaths(chatignore_path='./.aicastle/chat/.chatignore'):
    patterns = get_patterns(chatignore_path)
    return get_target_filepaths(patterns, True)


def get_finetuning_filepaths(finetuning_path='./.aicastle/chat/.finetuning'):
    patterns = get_patterns(finetuning_path)
    return get_target_filepaths(patterns, False)




import threading
import time

########### Thread ###########
def run_thread(func, *args, _start=True, **kwargs):
    """
    주어진 함수를 백그라운드 스레드에서 실행하고, 결과를 result_container에 저장합니다.
    
    :param func: 백그라운드에서 실행할 함수
    :param kwargs: 함수에 전달할 키워드 인수
    :return: 함수 실행 결과와 상태를 담는 딕셔너리
    """
    
    result_container = {'done': False, 'result': None, 'error': None}
    
    def wrapper():
        try:
            result = func(*args, **kwargs)
            result_container['result'] = result
        except Exception as error:
            result_container['error'] = error
        finally:
            result_container['done'] = True

    # 백그라운드에서 실행할 스레드 생성
    thread = threading.Thread(target=wrapper)
    result_container['thread'] = thread
    
    # 스레드를 데몬으로 설정 (메인 프로그램 종료 시 강제 종료)
    thread.daemon = True
    
    # 스레드 시작
    if _start :
        thread.start()
    
    return result_container


def run_thread_multi(func, num_workers=100, args_list=None, kwargs_list=None, _sleep=0.05):
    """
    동일한 함수를 여러 인자로 백그라운드 스레드에서 여러 번 실행하고, 결과를 저장합니다.
    
    :param func: 백그라운드에서 실행할 함수
    :param kwargs_list: 각 호출에 사용할 키워드 인수들의 리스트
    :param sleep: 각 스레드의 상태를 확인할 대기 시간 (초)
    :return: 각 호출의 결과와 상태를 담는 딕셔너리
    """
    
    if (args_list is None) and (kwargs_list is None):
        raise Exception("Either args_list or kwargs_list must be provided.")
    elif isinstance(args_list, list) and isinstance(kwargs_list, list) :
        if not len(args_list) == len(kwargs_list):
            raise Exception("The lengths of args_list and kwargs_list must be equal.")
        if len(args_list) == 0 :
            raise Exception("The provided args_list and kwargs_list cannot be empty.")
    elif kwargs_list is None :
        if len(args_list) == 0 :
            raise Exception("The provided args_list cannot be empty.")
        else :
            kwargs_list = [{} for _ in range(len(args_list))]
    elif args_list is None :
        if len(kwargs_list) == 0 :
            raise Exception("The provided kwargs_list cannot be empty.")
        else :
            args_list = [[] for _ in range(len(kwargs_list))]
    
    result_container = {'done': False, 'result': None, 'error': None, 'done_count':0, 'done_rate':0}

    def thread_control(result, num_workers = num_workers):
        result_len = len(result)
        working_idx_list = []
        for target_idx in range(result_len):
            while True :
                for working_idx in working_idx_list:
                    working_item = result[working_idx]
                    if working_item['done'] :
                        working_idx_list.remove(working_idx)
                if len(working_idx_list) < num_workers :
                    target_item = result[target_idx]
                    target_item['thread'].start()
                    working_idx_list.append(target_idx)
                    break

    def wrapper():
        try:
            result = []
            for args, kwargs in zip(args_list, kwargs_list):
                result.append(run_thread(func, *args, _start=False, **kwargs))
            result_container['result'] = result

            run_thread(thread_control, result)
            
            while True:
                result_len = len(result)
                done_count = 0
                for a_result in result:
                    if a_result['done']:
                        done_count +=1
                        
                if done_count == result_len:
                    result_container['done'] = True
                    result_container['done_count'] = done_count
                    result_container['done_rate'] = 1
                    break
                else :
                    result_container['done_count'] = done_count
                    result_container['done_rate'] = done_count / result_len
                     
                time.sleep(_sleep)
                
        except Exception as error:
            result_container['error'] = error
        finally:
            result_container['done'] = True
   
    # 백그라운드에서 실행할 스레드 생성
    thread = threading.Thread(target=wrapper)
    result_container['thread'] = thread
        
    # 스레드를 데몬으로 설정 (메인 프로그램 종료 시 강제 종료)
    thread.daemon = True
    
    # 스레드 시작
    thread.start()
    
    return result_container


def thread_join(result_container, max_time=60, sleep=0.05):
    """
    주어진 시간 내에 백그라운드 스레드가 완료될 때까지 대기합니다.
    
    :param result_container: 스레드의 결과와 상태를 담고 있는 딕셔너리
    :param max_time: 최대 대기 시간 (초)
    :param sleep: 각 스레드의 상태를 확인할 대기 시간 (초)
    :return: 완료된 스레드의 결과와 상태를 담은 딕셔너리
    """
    
    s_time = time.time()
    while not result_container['done']:
        if time.time() - s_time > max_time:
            raise Exception(f"max_time : {max_time} sec")
        time.sleep(sleep)
    return result_container



system_text = "You are the workspace chatbot of aicastle."

assistant_guidelines = [
"""

""".strip(),
]

finetuning_folder_path = ".aicastle/chat/finetuning"

model = "gpt-4o"
max_tokens = 16000
chat_token_price = 0.0035       # 1개 토큰 당 원화
finetuning_token_price = 0.035  # 1개 토큰 당 원화

max_thread_create_len = 32
message_create_workers = 20


from dotenv import load_dotenv
import streamlit as st

from aicastle.chat.utils import get_chat_file_hashes, load_system_text
from aicastle.chat.content_manager import OpenAIContentManager
from aicastle.chat.utils import load_config
import aicastle.chat.function_call as fc
from aicastle.chat.client import get_client, OpenAIChatManager
from aicastle.chat.tokens import (
    calculate_openai_messages_tokens_price,
    calculate_openai_finetuning_messages_tokens_price
)
load_dotenv(
    dotenv_path='.aicastle/chat/.env', 
    override=True  # 환경변수 덮어쓰기
)

config_data = load_config()

####### streamlit
st.title("AI Castle. Workspace Chat")

if "chat_file_hashes" not in st.session_state:
    chat_file_hashes = get_chat_file_hashes()
    st.session_state["chat_file_hashes"] = chat_file_hashes
else :
    chat_file_hashes = st.session_state["chat_file_hashes"]

if "chat_manager" not in st.session_state:
    incontext_messages = []
    for filepath in chat_file_hashes :
        content_manager = OpenAIContentManager(filepath)
        incontext_messages += content_manager.get_incontext_messages(
            model = config_data["model"],
            max_tokens = config_data["incontext"]["max_tokens"],
            in_text = config_data["incontext"]["in_text"],
            in_image = config_data["incontext"]["in_image"],
            info_modified = False,
            resolution=config_data["incontext"]["resolution"],
        )
    tokens_count, total_price = calculate_openai_messages_tokens_price(incontext_messages)
    print(" ============ incontext_messages (only text) ============ ")
    print(f"tokens_count: {tokens_count} / total_price: {total_price}")

    chat_manager = OpenAIChatManager(
        platform = config_data["platform"],
        chat_type = config_data["chat_type"],
        model = config_data["model"],
        assistant_id=config_data["assistant_id"],
        assistant_update=config_data["assistant_update"],
        temperature = config_data["temperature"],
        top_p = config_data["top_p"],
        stream = config_data["stream"],
        api_version = config_data["api_version"],
        additional_system_text = load_system_text(),
        messages = incontext_messages,
        function_call=config_data["function_call"],
        function_module=fc,
        response=False
    )
    st.session_state["chat_manager"] = chat_manager
else :
    chat_manager = st.session_state["chat_manager"]

if "show_messages" not in st.session_state:
    show_messages = []
    st.session_state["show_messages"] = show_messages
else :
    show_messages = st.session_state["show_messages"]

for message in show_messages:
    with st.chat_message(message["role"]):
        st.markdown(message["content"][0]["text"])


if user_input := st.chat_input("What is up?"):
    ### 유저 입력
    user_message = {"role": "user", "content": [{"type": "text","text": user_input}]}
    show_messages.append(user_message)
    with st.chat_message("user"):
        st.markdown(user_input)
    with st.spinner('Processing...'):  # 스피너 시작
        ### 파일 변경 감지
        new_chat_file_hashes = get_chat_file_hashes()
        new_incontext_messages = []
        for filepath, file_hash in new_chat_file_hashes.items() :
            if (not filepath in chat_file_hashes) or (chat_file_hashes[filepath] != file_hash):
                chat_file_hashes[filepath] = file_hash
                content_manager = OpenAIContentManager(filepath)
                new_incontext_messages += content_manager.get_incontext_messages(
                    model = config_data["model"],
                    max_tokens = config_data["incontext"]["max_tokens"],
                    in_text = config_data["incontext"]["in_text"],
                    in_image = config_data["incontext"]["in_image"],
                    info_modified = True,
                )
        if new_incontext_messages :
            chat_manager.add_messages(new_incontext_messages)

        ### 유저 request
        result_container = chat_manager.create([user_message])

        ### assistant response
        with st.chat_message("assistant"):
            assistant_content_box = st.empty()
            assistant_content = ""
            while True :
                if result_container['done'] and (len(result_container['texts']) == 0):
                    break
                else :
                    try :
                        text = result_container["texts"].pop(0)
                        assistant_content += text
                        assistant_content_box.markdown(assistant_content)
                    except :
                        pass
            
            ### submit_tool
            while result_container['tool_call_need'] :
                assistant_content += "\n---\nfunction_calling ....\n\n---\n"
                assistant_content_box.markdown(assistant_content)
                result_container = chat_manager.submit_tool_outputs(result_container['run'])
                while True :
                    if result_container['done'] and (len(result_container['texts']) == 0):
                        break
                    else :
                        try :
                            text = result_container["texts"].pop(0)
                            assistant_content += text
                            assistant_content_box.markdown(assistant_content)
                        except :
                            pass
        
            assistant_message = {"role": "assistant", "content": [{"type": "text","text": assistant_content}]}
            show_messages.append(assistant_message)


import yaml

from aicastle.chat.filepaths import get_chat_filepaths
from aicastle.utils.hash import get_hash_file

def load_config(config_path='.aicastle/chat/config.yml'):
    with open(config_path, 'r', encoding='utf-8') as file:
        config_data = yaml.safe_load(file)
    return config_data

def get_chat_file_hashes():
    filepaths = get_chat_filepaths()
    return {filepath:get_hash_file(filepath) for filepath in filepaths}

def load_system_text(system_text_path='.aicastle/chat/system.txt'):
    with open(system_text_path, 'r', encoding='utf-8') as file:
        system_text = file.read()
    return system_text


from openai import OpenAI, AzureOpenAI
import inspect
import aicastle.chat.hparams as chat_hp
import aicastle.chat.function_call as fc
import os
from openai.types.beta.assistant_stream_event import ThreadMessageDelta,ThreadRunRequiresAction
from openai.types.beta.threads.text_delta_block import TextDeltaBlock
import json
from concurrent.futures import ThreadPoolExecutor
import threading

def get_client(platform, api_version=None):
    if platform == 'openai':
        client = OpenAI(
            # api_key = os.environ["OPENAI_API_KEY"]
        )
        return client
    elif platform == 'azure_openai':
        client = AzureOpenAI(
            api_version=api_version,
            # api_key=os.environ["AZURE_OPENAI_API_KEY"],
            # azure_endpoint=os.environ["AZURE_OPENAI_ENDPOINT"],
        )
        return client
    else :
        raise Exception(f"Invalid platform : {platform}")

def get_tools(function_module=fc):
    function_names = [attr for attr in dir(function_module) if callable(getattr(function_module, attr))]
    return [
        {
            "type": "function",
            "function": {
                "name": function_name,
                "description": inspect.getsource(getattr(function_module, function_name))
            }
        }
        for function_name in function_names
    ]

class OpenAIChatManager:
    def __init__(self, 
            platform="openai", 
            chat_type="assistant", 
            model="gpt-4o",
            assistant_id=None,
            assistant_update=False,
            temperature=1,
            top_p=1,
            stream=True,
            additional_system_text="",
            api_version=None, 
            messages=None,
            function_call=False,
            function_module=None,
            response=False,
        ):
        self.platform = platform
        self.chat_type = chat_type
        self.additional_system_text = additional_system_text if additional_system_text else ""
        self.stream = stream
        self.function_module = function_module
        self.client = get_client(self.platform, api_version)
        self.response = response
        self.response_list = []
        
        messages = [] if messages is None else messages.copy()
        if self.chat_type == "assistant":
            self.set_assistant(model, assistant_id, assistant_update, additional_system_text, messages, temperature, top_p, function_call, function_module)
        elif self.chat_type == "chat":
            self.set_chat(model, additional_system_text, messages, temperature, top_p, stream, function_call, function_module)

            

    def run_control(self, run, async_=True, result_container=None):
        if async_ :
            result_container = {"texts":[], "run":None, "tool_call_need":None, "done":False}
            async_thread = threading.Thread(target=self.run_control, kwargs={"run":run, "async_": False, "result_container":result_container})
            result_container["async_thread"] = async_thread
            async_thread.start()
            return result_container

        else :
            if result_container is None :
                result_container = {"texts":[], "run":None, "tool_call_need":None, "done":False}
            
            if self.chat_type == "assistant":
                if self.stream :
                    for event in run:
                        # self.response_list.append(event)
                        # 응답 메세지 출력
                        if isinstance(event, ThreadMessageDelta):
                            if isinstance(event.data.delta.content[0], TextDeltaBlock):
                                result_container["texts"].append(event.data.delta.content[0].text.value)
                    
                    # run
                    result_container["run"] = event.data

                    # tool_call_need
                    if isinstance(event, ThreadRunRequiresAction):
                        result_container["tool_call_need"] = True
                    else :
                        result_container["tool_call_need"] = False

                else :
                    # 응답 메세지 출력
                    messages = self.client.beta.threads.messages.list(thread_id=self.thread_kwargs["thread_id"], run_id=run.id)
                    # self.messages = messages
                    if messages.data :
                        result_container["texts"].append(messages.data[0].content[0].text.value)

                    # run
                    result_container["run"] = run

                    # tool_call_need
                    if run.status == 'requires_action':
                        result_container["tool_call_need"] = True
                    else :
                        result_container["tool_call_need"] = False
                
            elif self.chat_type == "chat":
                if self.stream :
                    # 응답 메세지 출력
                    is_collecting_function_args = False
                    tool_calls = []
                    texts_join = ""
                    for event in run:
                        if not event.choices :
                            break
                        else :
                            last_event = event

                        delta = event.choices[0].delta
                        finish_reason = event.choices[0].finish_reason
                        if delta.content:
                            text = delta.content
                            texts_join += text
                            result_container["texts"].append(text)

                        if delta.tool_calls:
                            is_collecting_function_args = True
                            tool_call = delta.tool_calls[0]
                            if tool_call.id :
                                tool_call_target = tool_call
                            if tool_call.function.arguments:
                                tool_call_target.function.arguments += tool_call.function.arguments
                                
                        if finish_reason == "tool_calls" and is_collecting_function_args:
                            tool_calls.append(tool_call_target)
                            is_collecting_function_args = False

                    self.chat_kwargs["messages"].append({
                        "role":"assistant",
                        **({"content": [{"type": "text", "text": texts_join}]} if texts_join else {}),
                        **({"tool_calls":[{"id":tool_call.id, "type":"function", "function":{"name":tool_call.function.name, "arguments":tool_call.function.arguments}} for tool_call in tool_calls]} if tool_calls else {})
                    })

                    # run
                    if tool_calls :
                        last_event.choices[0].delta.tool_calls = tool_calls
                    result_container["run"] = last_event

                    # tool_call_need
                    if tool_calls :
                        result_container["tool_call_need"] = True
                    else :
                        result_container["tool_call_need"] = False
                    
                else :
                    self.chat_kwargs["messages"].append(run.choices[0].message.dict())
                    # 응답 메세지 출력
                    message_content = run.choices[0].message.content
                    if message_content :
                        result_container["texts"].append(message_content)

                    # run
                    result_container["run"] = run

                    # tool_call_need
                    if run.choices[0].message.tool_calls :
                        result_container["tool_call_need"] = True
                    else :
                        result_container["tool_call_need"] = False

            if self.response :
                self.response_list.append(result_container["run"])
            result_container["done"] = True
            return result_container

    def submit_tool_outputs(self, run, async_=True):
        if self.chat_type == "assistant":
            tool_outputs = []
            tool_calls = run.required_action.submit_tool_outputs.tool_calls
            for tool in tool_calls:
                tool_function = getattr(self.function_module, tool.function.name)
                tool_kwargs = json.loads(tool.function.arguments)
                output = tool_function(**tool_kwargs)
                tool_outputs.append({"tool_call_id": tool.id, "output":str(output)})
                print(f"- tool.function.name: {tool.function.name}")
                print(f"- tool_kwargs: {tool_kwargs}")
                print(f"- output: {output}")

            if self.stream :
                run = self.client.beta.threads.runs.submit_tool_outputs(
                    tool_outputs=tool_outputs,
                    thread_id=self.thread_kwargs["thread_id"],
                    run_id=run.id,
                    stream=True
                )
            else :
                run = self.client.beta.threads.runs.submit_tool_outputs_and_poll(
                    tool_outputs=tool_outputs,
                    thread_id=self.thread_kwargs["thread_id"],
                    run_id=run.id
                )

        elif self.chat_type == "chat":
            tool_calls = run.choices[0].delta.tool_calls if self.stream else run.choices[0].message.tool_calls 
            self.tool_calls = tool_calls
            for tool in tool_calls:
                tool_function = getattr(self.function_module, tool.function.name)
                tool_kwargs = json.loads(tool.function.arguments)
                output = tool_function(**tool_kwargs)
                self.chat_kwargs["messages"].append({
                    "role": "tool",
                    "tool_call_id": tool.id,
                    "content": [{"type": "text", "text": str(output)}],
                })
                print(f"- tool.function.name: {tool.function.name}")
                print(f"- tool_kwargs: {tool_kwargs}")
                print(f"- output: {output}")
            run = self.client.chat.completions.create(
                **self.chat_kwargs
            )

        result_container = self.run_control(run, async_=async_)
        return result_container
    
    def create(self, additional_messages, async_=True):
        if self.chat_type == "assistant":
            create = self.client.beta.threads.runs.create if self.stream else self.client.beta.threads.runs.create_and_poll
            run = create(
                additional_messages=additional_messages, 
                **self.thread_kwargs
            )

        elif self.chat_type == "chat":
            self.chat_kwargs["messages"] += additional_messages
            run = self.client.chat.completions.create(
                **self.chat_kwargs
            )
        
        result_container = self.run_control(run, async_=async_)
        return result_container

    def add_messages(self, messages):
        if self.chat_type == "assistant":
            # 병렬 요청이라 순서가 뒤바뀔 수 있음
            with ThreadPoolExecutor(max_workers=chat_hp.message_create_workers) as executor:
                futures = [
                    executor.submit(
                        self.client.beta.threads.messages.create,  # Keep this function unchanged
                        thread_id=self.thread_id,
                        role=message["role"],
                        content=message["content"]
                    )
                    for message in messages
                ]  
            # 요청 처리 대기
            futures_results = [future.result() for future in futures]
        elif self.chat_type == "chat":
            self.chat_kwargs["messages"] += messages
        else :
            raise Exception(f"Invalid chat_type : {self.chat_type}")
        
    def set_assistant(self, model, assistant_id, assistant_update, additional_system_text, messages, temperature, top_p, function_call, function_module):
        assistant_kwargs = {
            "name":os.path.basename(os.getcwd()),
            "model":model,
            "instructions":chat_hp.system_text+"/n"+additional_system_text,
            "temperature":temperature,
            "top_p":top_p,
            **({"tools": get_tools(function_module)} if function_call and function_module and get_tools(function_module) else {})
        }
        ### assistant 생성
        if assistant_id is None :
            assistant_id_path = ".aicastle/chat/__aicastlecache__/assistant_id"
            try:
                with open(assistant_id_path, "r", encoding='utf-8') as f:
                    assistant_id = f.read().strip()
                assistant = self.client.beta.assistants.update(
                    assistant_id,
                    **assistant_kwargs
                )
            except :
                assistant = self.client.beta.assistants.create(
                    **assistant_kwargs
                )
                os.makedirs(os.path.dirname(assistant_id_path), exist_ok=True)
                with open(assistant_id_path, "w", encoding='utf-8') as f:
                    f.write(assistant.id)
        else :
            if assistant_update :
                assistant = self.client.beta.assistants.update(
                    assistant_id,
                    **assistant_kwargs
                )
            else :
                assistant = self.client.beta.assistants.retrieve(
                    assistant_id
                )

        ### thread 생성
        thread = self.client.beta.threads.create(
            messages = messages[:chat_hp.max_thread_create_len] # 최대 32개 까지 밖에 안됨.....
        )
        self.thread_id = thread.id

        # 남은 메세지 추가
        self.add_messages(messages[chat_hp.max_thread_create_len:])

        self.thread_kwargs = {
            "thread_id":thread.id,
            "assistant_id":assistant.id,
            **({"stream": True} if self.stream else {})
        }

    def set_chat(self, model, additional_system_text, messages, temperature, top_p, stream, function_call, function_module):
        self.chat_kwargs = {
            "model":model,
            "temperature":temperature,
            "top_p":top_p,
            "stream":stream,
            "store":True,
            **({"stream_options":{"include_usage":True}} if stream else {}),
            "messages": [
                {"role":"system", "content":[{"type":"text", "text":chat_hp.system_text+"/n"+additional_system_text}]},
            ] + messages,
            **({"tools": get_tools(function_module), "parallel_tool_calls": False} if function_call and function_module and get_tools(function_module) else {}),
        }



import json


def dict2json(dict_data, ensure_ascii=False, indent=2):
	json_str = json.dumps(dict_data, ensure_ascii=ensure_ascii, indent=indent)
	return json_str


def json2dict(json_str):
    dict_data = json.loads(json_str)
    return dict_data



import os
import json
from datetime import datetime, timezone
import aicastle.chat.hparams as chat_hp

default_finetuning_folder_path = chat_hp.finetuning_folder_path

def get_utc_now():
    utc_now = datetime.now(timezone.utc)
    return utc_now.strftime("%Y-%m-%dT%H-%M-%SZ")


def save_finetuning_data(finetuning_data, filename=None, save_folder=default_finetuning_folder_path):
    os.makedirs(save_folder, exist_ok=True)

    if filename is None:
        filename = f"{get_utc_now()}.jsonl"  # 기본 파일 이름

    file_path = os.path.join(save_folder, filename)
    with open(file_path, 'w', encoding='utf-8') as f:
        for example in finetuning_data:
            f.write(json.dumps(example, ensure_ascii=False) + "\n")

    return file_path



import boto3
import os
from datetime import datetime, timezone, timedelta
import json
import pytz
import subprocess
import platform
from pathlib import Path
import yaml
import paramiko
import pandas as pd
import time
from matplotlib import pyplot as plt
from IPython.display import display

from dotenv import load_dotenv
load_dotenv()

class DRfCAWSClient:
    def __init__(self, config=None):
        if config is None :
            with open('config.yml') as f:
                self.config = yaml.safe_load(f)
        else :
            self.config = config

        if self.config['ec2-user'] is None:
            self.config['ec2-user'] = "ubuntu"

        self.ec2 = boto3.client('ec2',
            aws_access_key_id=os.getenv("AWS_ACCESS_KEY_ID"),
            aws_secret_access_key=os.getenv('AWS_SECRET_ACCESS_KEY'),
            region_name=self.config['region']
        )

        self.s3 = boto3.client('s3',
            aws_access_key_id=os.getenv("AWS_ACCESS_KEY_ID"),
            aws_secret_access_key=os.getenv('AWS_SECRET_ACCESS_KEY'),
            region_name=self.config['region']
        )

        self.cloudtrail = boto3.client('cloudtrail',
            aws_access_key_id=os.getenv("AWS_ACCESS_KEY_ID"),
            aws_secret_access_key=os.getenv('AWS_SECRET_ACCESS_KEY'),
            region_name=self.config['region']
        )

        self.ssh = None
        self.remote_path = f"/home/{self.config['ec2-user']}/deepracer-for-cloud"
        self.view_url = "http://aws-deepracer-school.s3.us-east-1.amazonaws.com/DRfC-multi-view.html"
        
    def get_instance_state(self):
        response = self.ec2.describe_instances(InstanceIds=[self.config['instance-id']])
        return response['Reservations'][0]['Instances'][0]['State']['Name']

    def get_instance_public_ip(self):
        if self.get_instance_state() != 'running':
            raise Exception("Instance is not running.")
        response = self.ec2.describe_instances(InstanceIds=[self.config['instance-id']])
        return response['Reservations'][0]['Instances'][0]['PublicIpAddress']

    def instance_wait_until_running(self):
        waiter = self.ec2.get_waiter('instance_running')
        waiter.wait(InstanceIds=[self.config['instance-id']])

    def instance_wait_until_stopped(self):
        waiter = self.ec2.get_waiter('instance_stopped')
        waiter.wait(InstanceIds=[self.config['instance-id']])

    def instance_wait_until_terminated(self):
        waiter = self.ec2.get_waiter('instance_terminated')
        waiter.wait(InstanceIds=[self.config['instance-id']])

    def start_instance(self, wait=True):
        self.ec2.start_instances(InstanceIds=[self.config['instance-id']])
        if wait:
            self.instance_wait_until_running()
        return self.get_instance_state()

    def stop_instance(self, wait=True):
        self.ec2.stop_instances(InstanceIds=[self.config['instance-id']])
        if wait:
            self.instance_wait_until_stopped()
        return self.get_instance_state()

    def terminate_instance(self, wait=True):
        self.ec2.terminate_instances(InstanceIds=[self.config['instance-id']])
        if wait:
            self.instance_wait_until_terminated(self.config['instance-id'])
        return self.get_instance_state()

    def get_s3_bucket_url(self):
        return f"https://{self.config["region"]}.console.aws.amazon.com/s3/buckets/{self.config['s3-bucket']}"

    def get_cloudtrail_events(self, start_utc=None):
        end_time = datetime.now(timezone.utc)
        if start_utc is None:
            if self.config["start-utc"] is None:
                start_time = end_time - timedelta(days=90)
            else :
                start_time = datetime.strptime(self.config["start-utc"], "%Y-%m-%dT%H:%M:%SZ")
        else :
            start_time = datetime.strptime(self.config["start-utc"], "%Y-%m-%dT%H:%M:%SZ")

        # 이벤트 이름으로 필터
        events = []
        for event_name in ["StartInstances", "StopInstances"]:
            response = self.cloudtrail.lookup_events(
                LookupAttributes=[
                    {
                        'AttributeKey': 'EventName',
                        'AttributeValue': event_name
                    },
                ],
                StartTime=start_time,
                EndTime=end_time
            )
            events.extend(response.get('Events', []))

        return events

    def calculate_instance_running_time(self, events=None, start_utc=None, instance_id=None, unit="hour", _print=False, print_tz_name="Asia/Seoul"):
        if events is None:
            events = self.get_cloudtrail_events(start_utc)
        if instance_id is None:
            instance_id = self.config['instance-id']
        
        ############ events 필터링 ############
        filtered_events = []
        for event in events:
            cloudtrail_event = event['CloudTrailEvent']
            detail = json.loads(cloudtrail_event)
            request_params = detail.get('requestParameters', {})
            instances_set = request_params.get("instancesSet", {})
            items = instances_set.get("items", [])
            instance_ids = []
            for item in items:
                if "instanceId" in item:
                    instance_ids.append(item["instanceId"])

            if instance_id in instance_ids:
                filtered_events.append(event)
        filtered_events.sort(key=lambda x: x['EventTime'])
        
        ############ 총 실행 시간 계산 ############
        if _print:
            try:
                tz = pytz.timezone(print_tz_name)
            except Exception as e:
                tz = timezone.utc

        total_running_time = timedelta(0)
        last_start_time = None
        for event in filtered_events:
            event_time = event['EventTime']
            event_name = event['EventName']

            if event_name == "StartInstances":
                if last_start_time is None:
                    last_start_time = event_time
                    # 한국 시간대로 변환하여 출력
                    if _print:
                        print(f"Start at {event_time.astimezone(tz)}")
                else:
                    if _print:
                        print(f"연속된 StartInstances 이벤트 발견. 이전 Start at {last_start_time.astimezone(tz)}, 현재 Start at {event_time.astimezone(tz)}. 이전 Start를 무시하고 최신 Start로 갱신하지 않습니다.")
            elif event_name == "StopInstances":
                if last_start_time is not None:
                    if event_time > last_start_time:
                        running_duration = event_time - last_start_time
                        total_running_time += running_duration
                        last_start_time = None
                        if _print:
                            print(f"Stop at {event_time.astimezone(tz)}, Duration: {running_duration}")
                    else:
                        if _print:
                            print(f"StopInstances 이벤트의 시간이 StartInstances 이전입니다. Stop at {event_time.astimezone(tz)}, Start at {last_start_time.astimezone(tz)}. 무시합니다.")
                else:
                    if _print:
                        print(f"StopInstances 이벤트가 StartInstances 없이 발생했습니다. Stop at {event_time.astimezone(tz)}. 무시합니다.")

        # 인스턴스가 현재 실행 중인 경우, 마지막 Start부터 현재까지의 시간을 추가
        if last_start_time is not None:
            end_time = datetime.now(timezone.utc)
            running_duration = end_time - last_start_time
            if running_duration.total_seconds() > 0:
                total_running_time += running_duration
                if _print:
                    print(f"Instance is currently running. Adding running duration from {last_start_time.astimezone(tz)} to {end_time.astimezone(tz)}: {running_duration}")
            else:
                if _print:
                    print(f"현재 시간보다 StartInstances 이벤트 시간이 더 큽니다. Start at {last_start_time.astimezone(tz)}, End at {end_time.astimezone(tz)}. 무시합니다.")

        total_sec = total_running_time.total_seconds()
        if unit == "second":
            return total_sec
        elif unit == "minute":
            return total_sec / 60
        elif unit == "hour":
            return total_sec / 3600
        else:
            return total_running_time
    
    ######### ssh #########
    def get_ssh_connect_command(self, chmod=True): 
        if chmod:
            self.chmod_instance_key()
        public_ip = self.get_instance_public_ip()
        return (
            f"ssh -i {self.config['instance-key-path']} "
            f"-o StrictHostKeyChecking=no "
            f"-o UserKnownHostsFile=/dev/null "
            f"-o ServerAliveInterval=60 "  # 60초마다 서버에 Ping
            f"-o ServerAliveCountMax=3 "  # 3번 실패 시 종료
            f"{self.config['ec2-user'].strip()}@{public_ip}"
        )


    def chmod_instance_key(self):
        if not os.path.exists(self.config['instance-key-path']):
            raise Exception(f"key file '{self.config['instance-key-path']}' not found")
        
        os_type = platform.system().lower()
        if os_type in ["linux", "darwin"]:
            subprocess.run(f"chmod 400 {self.config['instance-key-path']}", shell=True)
        elif os_type == "windows":
            subprocess.run(f"icacls {self.config['instance-key-path']} /inheritance:r /grant:r %USERNAME%:R", shell=True)
        else :
            raise Exception(f"Unsupported OS type: {os_type}")

    def get_ssh(self, force=False, chmod=True):
        if (self.ssh is None) or force :
            if chmod:
                self.chmod_instance_key()

            self.ssh = paramiko.SSHClient()
            self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            self.ssh.connect(
                hostname = self.get_instance_public_ip(),
                username = self.config['ec2-user'],
                key_filename= self.config['instance-key-path']
            )

            # keep alive
            transport = self.ssh.get_transport()
            transport.set_keepalive(30)

            # sftp
            self.sftp = self.ssh.open_sftp()

        return self.ssh


    def ssh_command(self, command, prefix=True, timeout=None):
        ssh = self.get_ssh()
        command_prefix = (
            "source ~/.bashrc && "
            "source ~/deepracer-for-cloud/bin/activate.sh && "
        )
        command = command_prefix + command if prefix else command
        # print(command)
        stdin, stdout, stderr = ssh.exec_command(command, timeout=timeout)
        exit_code = stdout.channel.recv_exit_status()
        output = stdout.read().decode()
        error = stderr.read().decode()
        if exit_code != 0:
            raise Exception(f"{error}\n{output}")

        return output

    def get_model_name(self):
        output = self.ssh_command('''grep "^DR_LOCAL_S3_MODEL_PREFIX=" ~/deepracer-for-cloud/run.env | cut -d'=' -f2''', prefix=False)
        return output.strip()

    def ssh_start_xrog(self):
        try :
            self.ssh_command("source ~/deepracer-for-cloud/utils/start-xorg.sh", timeout=10)
        except :
            pass

    def upload_best_model(self, local=True):
        command = "dr-upload-model -fb"
        command += "L" if local else ""
        self.ssh_command(command)

    def upload_car_zip(self, local=True):
        command = "dr-upload-car-zip -f"
        command += "L" if local else ""
        self.ssh_command(command)

    def upload_best_model_car_zip(self, _print=True):
        model_name_origin = self.get_model_name()
        model_name_best = model_name_origin + '-1'
        model_name_temp = model_name_origin + '-1-temp'
        print(f"[Best Model Name] {model_name_best}") if _print else None
        print("Best model uploading...", end="") if _print else None
        try :
            self.upload_best_model(local=True)
            upload_model_success = True
            print("done") if _print else None
            
        except :
            upload_model_success = False
            print("failed") if _print else None

        if upload_model_success :
            origin_DR_LOCAL_S3_MODEL_PREFIX = self.get_main_env("DR_LOCAL_S3_MODEL_PREFIX")
            self.set_main_env(DR_LOCAL_S3_MODEL_PREFIX=model_name_temp)
            origin_DR_LOCAL_S3_PRETRAINED = self.get_main_env("DR_LOCAL_S3_PRETRAINED")
            self.set_main_env(DR_LOCAL_S3_PRETRAINED=True)
            origin_DR_LOCAL_S3_PRETRAINED_PREFIX = self.get_main_env("DR_LOCAL_S3_PRETRAINED_PREFIX")
            self.set_main_env(DR_LOCAL_S3_PRETRAINED_PREFIX=model_name_best)
            origin_DR_WORKERS = self.get_main_env("DR_WORKERS")
            self.set_main_env(DR_WORKERS=1)

            prev_step_success = True
            if prev_step_success :
                try :
                    print("Start Sagemaker...", end="") if _print else None
                    self.start_training(wipes=True, retries=3, _print=False, suffix_force=True)
                    print("done") if _print else None
                except :
                    print("failed") if _print else None
                    prev_step_success = False

            self.set_main_env(force=True, DR_LOCAL_S3_MODEL_PREFIX=origin_DR_LOCAL_S3_MODEL_PREFIX)
            self.set_main_env(force=True, DR_LOCAL_S3_PRETRAINED=origin_DR_LOCAL_S3_PRETRAINED)
            self.set_main_env(force=True, DR_LOCAL_S3_PRETRAINED_PREFIX=origin_DR_LOCAL_S3_PRETRAINED_PREFIX)
            self.set_main_env(force=True, DR_WORKERS=origin_DR_WORKERS)

            if prev_step_success :
                try :
                    print("Upload car zip...", end="") if _print else None
                    self.upload_car_zip(local=True)
                    print("done") if _print else None
                except :
                    print("failed") if _print else None
                    prev_step_success = False

            try :
                print("Stop Sagemaker...", end="") if _print else None
                self.ssh_command("dr-stop-training")
                print("done") if _print else None
            except :
                print("failed") if _print else None

            self.s3.delete_object(Bucket=self.config['s3-bucket'], Key=f"{model_name_temp}/")
            best_model_s3_url = f"https://console.aws.amazonaws.com/s3/buckets/{self.config['s3-bucket']}/{model_name_best}/"
            print(f"Finished best model uploading: {best_model_s3_url}") if _print else None
        else :
            print("Skip best model uploading car zip.") if _print else None

    def get_train_viewer_url(self):
        if not self.is_training():
            raise Exception("Training is not running.")
        return self.view_url + f"?ip={self.get_instance_public_ip()}"

    def get_eval_viewer_url(self):
        if not self.is_evaluating():
            raise Exception("Evaluation is not running.")
        docker_ps_info = self.get_docker_ps_info()
        eval_info = docker_ps_info["eval"][0]
        return f"http://{self.get_instance_public_ip()}:{eval_info['port']}/stream_viewer?topic=/racecar/deepracer/kvs_stream"

    def is_training(self):
        docker_ps_info = self.get_docker_ps_info()
        if (len(docker_ps_info["train"]) > 0) or (len(docker_ps_info["rl_coach"]) > 0) or (len(docker_ps_info["sagemaker"]) > 0):
            return True
        else :
            return False
        
    def is_evaluating(self):
        docker_ps_info = self.get_docker_ps_info()
        if len(docker_ps_info["eval"]) > 0:
            return True
        else :
            return False
        
    def start_evaluation(self, _print=True):
        if self.is_evaluating():
            raise Exception("Evaluation is already running. Stop Evaluation first.")
        if self.is_training():
            raise Exception("Training is running. Stop Training first.")
        
        print(f"[Model Name] {self.get_model_name()}") if _print else None

        if self.get_main_env("DR_HOST_X") == "True":
            print("start xrog...", end="") if _print else None
            self.ssh_start_xrog()
            print("done") if _print else None

        success_keyword = "agent: Starting evaluation phase"
        timeout = 120
        command = (
            "dr-stop-evaluation && "
            "dr-start-evaluation"
        )
        output = self.ssh_command_pty(command, success_keyword, timeout)
        print(f"Evaluation Viewer: {self.get_eval_viewer_url()}") if _print else None
        
    def stop_evaluation(self, _print=True):
        if not self.is_evaluating():
            print("Evaluation is not running.") if _print else None
            return
        print(f"[Model Name] {self.get_model_name()}") if _print else None
        print("Stopping evaluation...") if _print else None
        self.ssh_command("dr-stop-evaluation")
        print("Evaluation stopped.") if _print else None


    def start_training(self, wipes=False, retries=2, _print=True, suffix_force=False):
        if self.is_training():
            raise Exception("Training is already running. Stop training first.")
        if self.is_evaluating():
            raise Exception("Evaluation is running. Stop Evaluation first.")
        
        model_name = self.get_model_name().strip()
        
        if not suffix_force :
            if model_name[-2:] == "-1":
                raise Exception((
                    f"Model name '{model_name}' is not allowed. "
                    "Model name must not end with '-1'. "
                    "Please change the model name."
                ))

        print(f"[Model Name] {model_name}") if _print else None

        if self.get_main_env("DR_HOST_X") == "True":
            print("start xrog...", end="") if _print else None
            self.ssh_start_xrog()
            print("done") if _print else None

        command = (
            "dr-stop-training && "
            "dr-update && "
            "dr-upload-custom-files && "
            "dr-start-training"
        )
        success_keyword = "DoorMan: installing SIGINT, SIGTERM"
        timeout = 60 * 5  # 5분

        dr_workers_env = int(drfc.get_main_env("DR_WORKERS"))
        for i in range(retries):
            try:
                print(f"Trying to start training ({i+1})...", end="") if _print else None
                output = self.ssh_command_pty(command + (" -w" if wipes else ""), success_keyword, timeout)
                dr_workers_real = len(drfc.get_docker_ps_info()['train'])
                
                if dr_workers_real != dr_workers_env:
                    raise Exception(f"DR_WORKERS is not matched. Expected: {dr_workers_env}, Real: {dr_workers_real}")
                
                # 성공하면 루프 탈출
                print("done") if _print else None
                break
            except Exception as e:
                print("failed")
                if "use -w option" in str(e):
                    raise Exception((
                        f"{self.get_model_name()} already exists.\n"
                        f"Use wipes option or delete the model from S3 bucket : {self.get_s3_bucket_url()}"
                    ))
                
                if i == retries - 1:
                    # 마지막 시도 실패
                    self.stop_training(_print=False)
                    print(f"Failed to start training after {retries} retries.") if _print else None
                    
                    raise e
                
                print(f"Retrying...") if _print else None
        
        print(f"Training Viewer : {self.get_train_viewer_url()}") if _print else None
        print(f"Model S3 URL : {self.get_s3_bucket_url()}/{model_name}/") if _print else None

    def stop_training(self, upload_best=True, _print=True):
        model_name_origin = self.get_model_name()
        print(f"[Model Name] {model_name_origin}") if _print else None

        if not self.is_training():
            print("Training is not running.") if _print else None
        else :
            print("Stopping training...", end="") if _print else None
            self.ssh_command("dr-stop-training")
            print("done") if _print else None

        if upload_best:
            self.upload_best_model_car_zip(_print=_print)


    def ssh_command_pty(self, command, success_keyword, timeout=60, prefix=True):
        ssh = self.get_ssh()
        command_prefix = (
            "source ~/.bashrc && "
            "source ~/deepracer-for-cloud/bin/activate.sh && "
        )
        command = command_prefix + command if prefix else command
        # print(command)
        stdin, stdout, stderr = ssh.exec_command(command, get_pty=True)

        channel = stdout.channel
        start_time = time.time()
        last_line = ""  # 가장 마지막으로 본 라인
        while True:
            # 1) stdout에 새 데이터가 들어왔는지 확인
            if channel.recv_ready():
                data = channel.recv(4096)  # 바이트로 받음
                data_decoded = data.decode('utf-8', errors='replace')  # 디코딩
                lines = data_decoded.split('\n')

                for line in lines:
                    if line.strip():
                        last_line = line
                    # 'Success' 키워드가 있으면 즉시 정상 종료
                    if success_keyword in line:
                        # 채널 닫아서 프로세스도 종료 시도 (tail -f 등은 강제 종료될 수 있음)
                        channel.close()
                        return last_line

            # 2) stderr에 새 데이터가 들어왔는지 확인
            #    여기선 "조금이라도 stderr가 찍히면 에러"로 판단
            if channel.recv_stderr_ready():
                err_data = channel.recv_stderr(4096)
                err_decoded = err_data.decode('utf-8', errors='replace')
                lines = err_decoded.split('\n')

                # 마지막 줄 갱신
                for line in lines:
                    if line.strip():
                        last_line = line

                # stderr에 뭔가 찍혔으니 곧바로 에러로 처리
                channel.close()
                raise Exception(last_line)
            
            # 3) 프로세스가 종료되었는지 (exit_status_ready) 확인
            #    정상(0) 종료면 True 리턴, 비정상 종료면 에러
            if channel.exit_status_ready():
                exit_code = channel.recv_exit_status()
                if exit_code == 0:
                    return last_line
                else:
                    raise Exception(last_line)

            if time.time() - start_time > timeout:
                channel.close()
                raise Exception(last_line)
            
            # 0.1초 대기 (CPU 과부하 방지)
            time.sleep(0.1)
        
    def ssh_sed(self, target_path, default_path=None, copy=False, **kwargs): 
        command = ""
        if copy:
            command += f"cp {default_path} {target_path}" + "\n"
        for key, value in kwargs.items():
            command += f'''sed -i \'s/\\(^{key}=\\).*/\\1{value}/\' {target_path}''' + "\n"

        self.ssh_command(command, prefix=False)

    def get_main_env(self, variable_name):
        system_output = drfc.ssh_command(f"grep '^{variable_name}=' ~/deepracer-for-cloud/system.env | cut -d '=' -f 2-", prefix=False).strip()
        run_output = drfc.ssh_command(f"grep '^{variable_name}=' ~/deepracer-for-cloud/run.env | cut -d '=' -f 2-", prefix=False).strip()
        env_value = system_output or run_output or None
        return env_value
    
    def set_main_env(self, copy=False, force=False, **kwargs):
        if not force :
            if self.is_training():
                raise Exception("Training is running. Stop Training first.")
            if self.is_evaluating():
                raise Exception("Evaluation is running. Stop Evaluation first.")
        
        # system.env
        env_file_path = os.path.join(self.remote_path, 'system.env')
        env_default_file_path = os.path.join(self.remote_path, 'defaults', 'template-system.env')
        self.ssh_sed(env_file_path, env_default_file_path, False, **kwargs)
        # run.env
        env_file_path = os.path.join(self.remote_path, 'run.env')
        env_default_file_path = os.path.join(self.remote_path, 'defaults', 'template-run.env')
        self.ssh_sed(env_file_path, env_default_file_path, copy, **kwargs)

    def set_subworker_env(self, worker_id, copy=False, force=False, **kwargs):
        if not force :
            if self.is_training():
                raise Exception("Training is running. Stop Training first.")
            if self.is_evaluating():
                raise Exception("Evaluation is running. Stop Evaluation first.")
        
        if not ((worker_id > 1) and isinstance(worker_id, int)):
            raise Exception("worker_id must be 2 or more int")
        
        env_file_path = os.path.join(self.remote_path, f'worker-{worker_id}.env')
        # env_default_file_path = os.path.join(self.remote_path, 'defaults', 'template-worker.env')
        env_default_file_path = os.path.join(self.remote_path, 'run.env')
        self.ssh_sed(env_file_path, env_default_file_path, copy, **kwargs)

    def ssh_echo(self, string, remote_file_path):
        command = f"echo '{string}' > {remote_file_path}"
        self.ssh_command(command, prefix=False)

    def set_reward_function(self, file_path="reward_function.py", force=False):
        if not force:
            if self.is_training():
                raise Exception("Training is running. Stop Training first.")
            if self.is_evaluating():
                raise Exception("Evaluation is running. Stop Evaluation first.")

        with open(file_path, 'r') as f:
            reward_function_str = f.read()

        remote_file_path = os.path.join(self.remote_path, "custom_files", "reward_function.py")
        command = f"cat > {remote_file_path} << EOF\n{reward_function_str}\nEOF"
        self.ssh_command(command, prefix=False)

    def set_hyperparameters(self, dict_data, force=False):
        if not force:
            if self.is_training():
                raise Exception("Training is running. Stop Training first.")
            if self.is_evaluating():
                raise Exception("Evaluation is running. Stop Evaluation first.")
        
        remote_file_path = os.path.join(self.remote_path, "custom_files", "hyperparameters.json")
        self.ssh_echo(json.dumps(dict_data), remote_file_path)

    def set_model_metadata(self, dict_data, force=False):
        if not force :
            if self.is_training():
                raise Exception("Training is running. Stop Training first.")
            if self.is_evaluating():
                raise Exception("Evaluation is running. Stop Evaluation first.")
        
        remote_file_path = os.path.join(self.remote_path, "custom_files", "model_metadata.json")
        self.ssh_echo(json.dumps(dict_data), remote_file_path)

    def cat_run_env(self):
        command = "cat ~/deepracer-for-cloud/run.env"
        print(self.ssh_command(command))

    def cat_system_env(self):
        command = "cat ~/deepracer-for-cloud/system.env"
        print(self.ssh_command(command))

    def cat_worker_env(self, worker_id):
        if not ((worker_id > 1) and isinstance(worker_id, int)):
            raise Exception("worker_id must be 2 or more int")
        command = f"cat ~/deepracer-for-cloud/worker-{worker_id}.env"
        print(self.ssh_command(command))

    def cat_hyperparameters(self):
        # json 파일을 불러와서 dict로 print
        command = "cat ~/deepracer-for-cloud/custom_files/hyperparameters.json"
        output = self.ssh_command(command)
        # print 할 때 예쁘게
        print(json.dumps(json.loads(output), indent=4))

    def cat_model_metadata(self):
        # json 파일을 불러와서 dict로 print
        command = "cat ~/deepracer-for-cloud/custom_files/model_metadata.json"
        output = self.ssh_command(command)
        print(json.dumps(json.loads(output), indent=4))
    
    def cat_reward_function(self):
        command = "cat ~/deepracer-for-cloud/custom_files/reward_function.py"
        print(self.ssh_command(command))


    def get_df_docker_ps(self):
        command = "docker ps --format '{{.ID}};{{.Image}};{{.Command}};{{.CreatedAt}};{{.Status}};{{.Ports}};{{.Names}}'"
        output = self.ssh_command(command, prefix=False)
        lines = output.strip().split('\n')
        data = [line.split(';') for line in lines if line.strip()]

        columns = ['ID', 'Image', 'Command', 'CreatedAt', 'Status', 'Ports', 'Names']
        df = pd.DataFrame(data, columns=columns)
        return df
    
    def get_docker_ps_info(self):
        df_docker_ps = self.get_df_docker_ps()
        docker_ps_info = {"train":[], "eval":[], "rl_coach":[], "sagemaker":[], "viewer":[], "analysis":[], "unknown":[]}
        df_len = df_docker_ps.shape[0]
        for i in range(df_len):
            row = df_docker_ps.iloc[i]
            port = None
            index = None
            if '-robomaker-' in row['Names']:
                if 'eval' in row['Names']:
                    container_type = 'eval'
                else :
                    container_type = 'train'
                port = row['Ports'].split('->8080/tcp')[0].split(':')[-1]
                index = row['Names'].split('-')[-1]
            elif '-rl_coach-' in row['Names']:
                container_type = 'rl_coach'
            elif '-algo-' in row['Names']:
                container_type = 'sagemaker'
            elif '-viewer-' in row['Names']:
                container_type = 'viewer'
                port = row['Ports'].split('->80/tcp')[0].split(':')[-1]
            elif 'deepracer-analysis' in row['Names']:
                container_type = 'analysis'
            else :
                container_type = 'unknown'

            docker_ps_info[container_type].append({"index":index, "port":port, "row":row})

        return docker_ps_info

    ######### mount #########
    def mount_install(self):
        os_type = platform.system().lower()
        if os_type == "linux":
            # SSHFS 설치 여부 확인
            result = subprocess.run(["which", "sshfs"], capture_output=True, text=True)
            if result.returncode != 0:
                print("Installing SSHFS...")
                subprocess.run(["sudo", "apt-get", "update"], check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
                subprocess.run(["sudo", "apt-get", "install", "-y", "sshfs"], check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        elif os_type == "darwin":
            # macOS에서 Homebrew를 사용하여 SSHFS 설치 여부 확인
            result = subprocess.run(["brew", "list", "sshfs"], capture_output=True, text=True)
            if "Not installed" in result.stderr or result.returncode != 0:
                print("Installing SSHFS...")
                subprocess.run(["brew", "update"], check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
                subprocess.run(["brew", "install", "sshfs"], check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
        else :
            raise Exception(f"Unsupported OS type: {os_type}")

    def is_mounted(self, mount_path='./mnt/deepracer-for-cloud'):
        mount_path = os.path.abspath(mount_path)
        result = subprocess.run(["mountpoint", "-q", mount_path])
        return result.returncode == 0

    def mount(self, mount_path='./mnt/deepracer-for-cloud', chmode=True):
        self.mount_install()
        mount_path = os.path.abspath(mount_path)
        Path(mount_path).mkdir(parents=True, exist_ok=True)

        public_ip = self.get_instance_public_ip()
        if chmode:
            self.chmod_instance_key()

        # 이미 마운트되어 있는지 확인
        if self.is_mounted(mount_path):
            print(f"Already mounted to '{mount_path}'.")
            return
        else:
            print(f"Mounting to '{mount_path}'...")

        # SSHFS를 사용하여 마운트
        sshfs_command = [
            "sshfs",
            "-o", "StrictHostKeyChecking=no",         # 호스트 키 확인 생략
            "-o", "UserKnownHostsFile=/dev/null",     # 호스트 키 저장 안함
            "-o", "reconnect",                        # 재연결 시도
            "-o", "ServerAliveInterval=15",           # 서버에 주기적인 패킷 전송
            "-o", "ServerAliveCountMax=3",            # 서버 응답 없는 경우 재연결 횟수
            "-o", f'''IdentityFile="{os.path.abspath(self.config["instance-key-path"])}"''',       # SSH 키 파일 지정
            f"{self.config['ec2-user'].strip()}@{public_ip}:{self.remote_path}",   # 원격 경로
            mount_path                             # 로컬 마운트 포인트
        ]

        # SSHFS 명령어를 백그라운드에서 실행하도록 Popen 사용
        process = subprocess.Popen(
            sshfs_command,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.PIPE,
            preexec_fn=os.setsid
        )

        # SSHFS가 제대로 마운트되었는지 확인
        stdout, stderr = process.communicate(timeout=30)  # 30초 내에 마운트 완료 여부 확인

        if process.returncode == 0:
            print(f"Successfully mounted '{public_ip}:{self.remote_path}' to '{mount_path}'.")
        else:
            print(f"Failed to mount '{public_ip}:{self.remote_path}' to '{mount_path}'.")
            print("Error Output:")
            print(stderr.decode())
            raise subprocess.CalledProcessError(
                returncode=process.returncode,
                cmd=sshfs_command,
                output=stdout,
                stderr=stderr
            )

        # subprocess.run(sshfs_command, shell=True, check=True)
        # print(f"Successfully mounted '{public_ip}:{remote_dir}' to '{local_dir}'.")

    def unmout(self, mount_path='./mnt/deepracer-for-cloud'):
        mount_path = os.path.abspath(mount_path)
        if self.is_mounted(mount_path):
            subprocess.run(["umount", mount_path], check=True)
            print(f"Successfully unmounted '{mount_path}'.")
        else:
            print(f"'{mount_path}' is not mounted.")

    ###### metrics ######
    def get_metrics_key_list(self, model_name=None, target=None):
        if model_name is None:
            model_name = self.get_model_name()
        
        response = self.s3.list_objects_v2(
            Bucket=self.config['s3-bucket'],
            Prefix=os.path.join(model_name, "metrics")
        )

        if "Contents" in response:
            metrics_key_list = [content["Key"] for content in response["Contents"]]
        else :
            metrics_key_list = []

        if target == "training":
            metrics_key_list = [key for key in metrics_key_list if "TrainingMetrics" in key]
        elif target == "evaluation":
            metrics_key_list = [key for key in metrics_key_list if "evaluation" in key]

        metrics_key_list.sort()
        return metrics_key_list

    def show_training_metrics_from_key(self, metrics_key, alpha=0.15, ma_window=10):
        data = self.s3.get_object(
            Bucket=self.config['s3-bucket'],
            Key=metrics_key
        )["Body"].read().decode("utf-8")
        data_dict = json.loads(data)
        df_metrics = pd.DataFrame(data_dict['metrics'])
        df_training = df_metrics[df_metrics['phase'] == 'training'].copy()
        df_evaluation = df_metrics[df_metrics['phase'] == 'evaluation'].copy()
        df_training['reward_score_ma'] = df_training['reward_score'].rolling(window=ma_window).mean()
        df_training['completion_percentage_ma'] = df_training['completion_percentage'].rolling(window=ma_window).mean()
        df_evaluation_group = df_evaluation.groupby('episode')

        # x축의 전체 범위 계산
        x_min = 0
        x_max = max(df_training['episode'].max(), df_evaluation['episode'].max())

        # Subplots 생성 (공유 x축 사용)
        fig, axes = plt.subplots(2, 1, figsize=(10, 10), sharex=True, constrained_layout=True)

        # reward score plot
        axes[0].scatter(df_training['episode'], df_training['reward_score'], alpha=alpha, c='green')
        axes[0].plot(df_training['episode'], df_training['reward_score_ma'], c='green', label='training')
        axes[0].scatter(df_evaluation['episode'], df_evaluation['reward_score'], alpha=alpha, c='magenta')
        axes[0].plot(df_evaluation_group['episode'].first(), df_evaluation_group['reward_score'].mean(), c='magenta', label='evaluation')
        axes[0].legend()
        axes[0].set_ylabel('reward')
        axes[0].set_title('Reward Score')
        axes[0].grid()

        # completion percentage plot
        axes[1].scatter(df_training['episode'], df_training['completion_percentage'], alpha=alpha, c='blue')
        axes[1].plot(df_training['episode'], df_training['completion_percentage_ma'], c='blue', label='training')
        axes[1].scatter(df_evaluation['episode'], df_evaluation['completion_percentage'], alpha=alpha, c='red')
        axes[1].plot(df_evaluation_group['episode'].first(), df_evaluation_group['completion_percentage'].mean(), c='red', label='evaluation')
        axes[1].legend()
        axes[1].set_ylim(0, 100)  # completion percentage는 0~100 범위로 고정
        axes[1].set_xlabel('episode')
        axes[1].set_ylabel('completion %')
        axes[1].set_title('Completion Percentage')
        axes[1].grid()

        # x축 범위 및 공유 설정
        axes[0].set_xlim(x_min, x_max)

        # 전체 플롯 표시
        plt.show()

    def show_training_metrics(self, model_name=None, worker_ids=None, alpha=0.07, ma_window=10):
        if model_name is None:
            model_name = self.get_model_name()
        print(f"[Model Name] {model_name}")

        if worker_ids is None:
            worker_ids = [1]
        elif isinstance(worker_ids, int):
            worker_ids = [worker_ids]
        elif isinstance(worker_ids, list):
            pass
        else:
            raise Exception("worker_ids must be int or list of int.")
        
        metrics_key_list = self.get_metrics_key_list(model_name, target="training")
        for worker_id in worker_ids:
            metrics_idx = worker_id - 1
            metrics_suffix = f"_{metrics_idx}" if metrics_idx != 0 else ""
            metrics_key = f"{model_name}/metrics/TrainingMetrics{metrics_suffix}.json"
            if metrics_key in metrics_key_list:
                print(f"Worker {worker_id} : {metrics_key}")
                self.show_training_metrics_from_key(metrics_key, alpha, ma_window)
            else :
                print(f"Worker {worker_id} : No metrics found.")

    def show_evaluation_metrics(self, model_name=None, job_idx=-1):
        if self.is_evaluating():
            raise Exception("Evaluation is running")
        
        if model_name is None:
            model_name = self.get_model_name()

        print(f"[Model Name] {model_name}")
        
        metrics_key_list = self.get_metrics_key_list(model_name, target="evaluation")
        if len(metrics_key_list) < 1:
            print("No evaluation metrics found.")
            return

        # df_metrics
        metrics_key = metrics_key_list[job_idx]
        data = self.s3.get_object(
            Bucket=self.config['s3-bucket'],
            Key=metrics_key
        )["Body"].read().decode("utf-8")
        data_dict = json.loads(data)
        df_metrics = pd.DataFrame(data_dict['metrics'])
        df_metrics['seconds'] = df_metrics['metric_time']/1000
        # df_metrics.set_index('trial', inplace=True)
        df_metrcis2 = df_metrics[['trial', 'seconds', 'off_track_count', 'crash_count', 'reset_count']]
        display(df_metrcis2)
        
        # Job name
        metrics_basename = os.path.basename(metrics_key)
        eval_job_name = metrics_basename.split(".")[0]
        print(f"Job Name: {eval_job_name}")

        # show video url
        try :
            presigned_url = self.s3.generate_presigned_url(
                'get_object',
                Params={'Bucket': self.config['s3-bucket'], 'Key': f"{model_name}/mp4/camera-pip/0-video.mp4"},
                ExpiresIn=3600
            )
            print(f"Video URL: {presigned_url}")
        except:
            print("Video not found.")
        

drfc = DRfCAWSClient()


from PIL import Image
import base64
from io import BytesIO
import os
import fitz  # PyMuPDF
from aicastle.chat.tokens import split_tokens
import aicastle.chat.hparams as chat_hp

default_system_text = chat_hp.system_text
default_assistant_guidelines = chat_hp.assistant_guidelines

class OpenAIContentManager:
    def __init__(self, target, img_max = 1024, dpi = 90):
        self.thumbnail_size = (img_max, img_max)
        self.dpi = dpi

        if isinstance(target, str):     # 파일 경로에서 가져오기 (ex) path/to/image.jpg
            self.data = self.loader(target)
            self.filepath = target
        elif isinstance(target, dict):  # 완성된 content data (ex) {"type":"image_url", "data":"https://..."}
            self.data = target
            self.filepath = None
        else :
            raise Exception(f"Invalid target : {type(target)}")
        
        self.type = self.data["type"] if self.data else None

    def get_incontext_messages(self, model="gpt-4o", resolution="auto", max_tokens=50000, in_text=True, in_image=False, info_modified=False):
        if info_modified :
            content = [
                {"type":"text", "text":"(info) 다음 파일이 수정 됨."}
            ]
        else :
            content = []
        
        if self.type == "text":
            if in_text :
                content += \
                    [{"type":"text", "text":f'''<<<context filepath="{self.filepath}">>>'''}] +\
                    [{"type":"text", "text":tokens} for tokens in split_tokens(self.data["data"], max_tokens, model=model)] +\
                    [{"type":"text", "text":f'''<<<///context>>>'''}]
            else :
                content.append({"type":"text", "text":f'''<<<context filepath="{self.filepath}"///>>>'''})
        elif self.type == "image_url":
            if in_image :
                content += [
                    {"type":"text", "text":f'''<<<context filepath="{self.filepath}">>>'''},
                    {"type":"image_url", "image_url":{"url":self.data["data"], "resolution":resolution}},
                    {"type":"text", "text":f'''<<<///context>>>'''}
                ]
            else :
                content.append({"type":"text", "text":f'''<<<context filepath="{self.filepath}"///>>>'''})
        elif self.type == "pdf_pages":
            if in_text or in_image :
                for page_data in self.data["data"]:
                    content.append({"type":"text", "text":f'''<<<context filepath="{self.filepath}" page_num="{page_data["page_num"]}">>>'''})
                    if in_text :
                        for tokens in split_tokens(page_data["text"], max_tokens, model=model):
                            content.append({"type":"text", "text":tokens})
                    if in_image :
                        content.append({"type":"image_url", "image_url":{"url":page_data["image_url"], "resolution":resolution}})
                    content.append({"type":"text", "text":f'''<<<///context>>>'''})
            else :
                content.append({"type":"text", "text":f'''<<<context filepath="{self.filepath}"///>>>'''})
        else :
            content.append({"type":"text", "text":f'''<<<context filepath="{self.filepath}"///>>>'''})
        
        messages = [{"role":"user", "content":content}]
        return messages
    
    def get_finetuning_data(self, system_text=default_system_text, max_tokens=16000, model="gpt-4o", in_text=True, in_image=False):
        system_message = {"role":"system", "content":[{"type":"text", "text":system_text}]}
        finetuning_data = []

        #### 가이드라인 학습

        #### 데이터 학습
        if self.type == "text":
            if in_text :
                messages = [
                    system_message,
                    {"role":"user", "content": [{"type":"text", "text":"파일경로 중 하나를 작성하시오."}]},
                    {"role":"assistant", "content": [{"type":"text", "text":self.filepath}]},
                ]
                tokens_list = split_tokens(self.data["data"], max_tokens, model=model)
                if len(tokens_list) > 0 :
                    messages.append({"role":"user", "content": [{"type":"text", "text":"이 파일의 내용을 작성하시오."}]})
                    messages += [{"role":"assistant", "content": [{"type":"text", "text":tokens}]} for tokens in tokens_list]
                
                finetuning_data.append({"messages":messages})
        elif self.type == "image_url" :
            if in_image :
                messages = [
                    system_message,
                    {"role":"user", "content": [{"type":"text", "text":"다음 이미지의 파일경로를 작성하시오."}, {"type":"image_url", "image_url":{"url":self.data["data"]}}]},
                    {"role":"assistant", "content": [{"type":"text", "text":self.filepath}]},
                ]
                finetuning_data.append({"messages":messages})
        elif self.type == "pdf_pages":
            ### 메세지 경로 맞추기
            messages = [
                system_message,
                {"role":"user", "content": [{"type":"text", "text":"파일경로 중 하나를 작성하시오."}]},
                {"role":"assistant", "content": [{"type":"text", "text":self.filepath}]},
                {"role":"user", "content": [{"type":"text", "text":"이 파일의 총 페이지 수를 작성하시오."}]},
                {"role":"assistant", "content": [{"type":"text", "text":str(len(self.data["data"]))}]},
            ]
            finetuning_data.append({"messages":messages})
            ### 페이지 내용 맞추기 (텍스트)
            if in_text :
                for page_data in self.data["data"]:
                    tokens_list = split_tokens(page_data["text"], max_tokens, model=model)
                    if len(tokens_list) > 0 :
                        messages = [
                            system_message,
                            {"role":"user", "content": [{"type":"text", "text":f"{self.filepath} 파일의 {page_data['page_num']} 페이지에서 추출된 텍스트를 작성하시오."}]},
                        ]
                        messages += [{"role":"assistant", "content": [{"type":"text", "text":tokens}]} for tokens in tokens_list]
                        finetuning_data.append({"messages":messages})
            ### 페이지 내용 맞추기 (이미지)
            if in_image :
                for page_data in self.data["data"]:
                    messages = [
                        system_message,
                        {"role":"user", "content": [{"type":"text", "text":"다음 이미지의 파일경로를 작성하시오."}, {"type":"image_url", "image_url":{"url":self.data["data"]}}]},
                        {"role":"assistant", "content": [{"type":"text", "text":f"{self.filepath} 파일의 {page_data['page_num']} 페이지"}]}
                    ]
                    finetuning_data.append({"messages":messages})
        return finetuning_data

    def loader(self, filepath):
        # 텍스트 파일
        try :
            with open(filepath, 'r', encoding='utf-8') as file:
                text = file.read()  # 파일 전체 읽기
                return {"type":"text", "data":text}
        except :
            pass

        # 이미지 파일
        try :
            with Image.open(filepath) as img:
                img.thumbnail(self.thumbnail_size)
                with BytesIO() as buffer:
                    img.save(buffer, format=img.format)
                    base64_str = base64.b64encode(buffer.getvalue()).decode('utf-8')
                    mime_type = f"image/{img.format.lower()}"
                    base64_str_url = f"data:{mime_type};base64,{base64_str}"
                    return {"type":"image_url", "data":base64_str_url}
        except :
            pass

        # pdf 파일
        if os.path.splitext(filepath)[-1].lower() == '.pdf':
            try : 
                pdf = fitz.open(filepath)
                pdf_pages = []
                for page_idx, page in enumerate(pdf) :
                    pixmap = page.get_pixmap(dpi=self.dpi)
                    pixmap_bytes = pixmap.tobytes()
                    page_image_buffer = BytesIO(pixmap_bytes)
                    with Image.open(page_image_buffer) as img:
                        with BytesIO() as buffer:
                            img.save(buffer, format="PNG")  # PDF 페이지는 PNG 형식으로 저장
                            base64_str = base64.b64encode(buffer.getvalue()).decode('utf-8')
                            mime_type = "image/png"
                            base64_str_url = f"data:{mime_type};base64,{base64_str}"
                    
                    pdf_pages.append({"page_num":page_idx + 1, "text":page.get_text(), "image_url":base64_str_url})
                pdf.close()
                return {"type":"pdf_pages", "data":pdf_pages}
            except :
                pass
        
        # 지원되지 않는 파일
        return None



import tiktoken
import aicastle.chat.hparams as chat_hp

default_model = chat_hp.model
default_max_tokens = chat_hp.max_tokens
default_chat_token_price = chat_hp.chat_token_price
default_finetuning_token_price = chat_hp.finetuning_token_price


def calculate_tokens(text: str, model: str = default_model, return_count=True):
    try:  # 모델에 맞는 토크나이저 가져오기
        encoding = tiktoken.encoding_for_model(model)
    except KeyError:  # 기본 토크나이저 (호환되지 않는 모델의 경우)
        encoding = tiktoken.get_encoding("cl100k_base")

    tokens = encoding.encode(text)
    if return_count :
        return len(tokens)
    else :
        return tokens


def split_tokens(text: str, max_tokens: int = default_max_tokens, model: str = default_model):
    tokens = calculate_tokens(text, model=model, return_count=False)
    encoding = tiktoken.encoding_for_model(model)
    chunks = [tokens[i:i + max_tokens] for i in range(0, len(tokens), max_tokens)]
    return [encoding.decode(chunk) for chunk in chunks]


def calculate_openai_messages_tokens_price(messages, model=default_model, token_price=default_chat_token_price):
    tokens_count = 0
    for message in messages :
        for content in message["content"]:
            if isinstance(content, str) :
                tokens_count += calculate_tokens(content, model, return_count=True)
            elif isinstance(content, dict):
                if content["type"] == "text":
                    tokens_count += calculate_tokens(content["text"], model, return_count=True)
                else :
                    # 이미지 토큰 계산 추후 업데이트
                    pass
            else :
                raise Exception(f"Invalid content type : {type(content)}")
    total_price = tokens_count * token_price
    return tokens_count, total_price


def calculate_openai_finetuning_messages_tokens_price(finetuning_data, model=default_model, token_price=default_finetuning_token_price):
    tokens_count = 0
    for messages in finetuning_data:
        for message in messages["messages"] :
            for content in message["content"]:
                if isinstance(content, str) :
                    tokens_count += calculate_tokens(content, model, return_count=True)
                elif isinstance(content, dict):
                    if content["type"] == "text":
                        tokens_count += calculate_tokens(content["text"], model, return_count=True)
                    else :
                        # 이미지 토큰 계산 추후 업데이트
                        pass
                else :
                    raise Exception(f"Invalid content type : {type(content)}")
    total_price = tokens_count * token_price
    return tokens_count, total_price


import pkg_resources
from PIL import Image
from io import BytesIO

class PackageImageLoader:
    def __init__(self, package_image_path):
        self.package_image_path = package_image_path
        self.image_path = self.get_image_path()
        # self.image = self.load_image()
    
    def get_image_path(self):
        # 패키지 내 데이터 파일 경로
        return pkg_resources.resource_filename(__name__, f'package_data/{self.package_image_path}')
    
    def load_image(self):
        # 이미지 열기
        with open(self.image_path, 'rb') as img_file:
            image_data = img_file.read()
        image = Image.open(BytesIO(image_data))
        return image


import os


def chdir_root(root_name = "app"):
	current_path = os.getcwd()
	origin_path = current_path
	current_path_basename = os.path.basename(current_path)

	if current_path_basename == root_name :
		return current_path

	if os.path.isdir(root_name):
		os.chdir(root_name)
		current_path = os.getcwd()
		return current_path

	while True :
		os.chdir("..")
		current_path = os.getcwd()
		current_path_basename = os.path.basename(current_path)
		if current_path_basename == root_name:
			return current_path
		if current_path_basename == "":
			os.chdir(origin_path)
			raise Exception("No path")


from setuptools import setup, find_packages
from get_version import read_version

setup(
    name='aicastle',
    version=read_version(),
    packages=find_packages(include=['aicastle', 'aicastle.*']),
    include_package_data=True,
    package_data={
        'aicastle': ['package_data/*'],
    },
    
    # 의존성
    install_requires=[ 
        # 'tqdm', 
        # 'pandas', 
        # 'scikit-learn',
        "click",      # CLI 도구

        ## drfc-aws
        "boto3",
        "paramiko",
        # "python-dotenv",

        ### chat
        "pyyaml",
        "streamlit",  # Streamlit 의존성
        'openai',
        'tiktoken',
        'pathspec',
        'pymupdf',
        'python-dotenv',
        # 'python-magic',
        
        # 'azure-identity',
        # 'boto3',

        'pillow',
    ],

    entry_points={
        "console_scripts": [
            "aicastle=aicastle.cli:main",  # aicastle 명령어 실행 엔트리포인트
        ],
    },

    author='aicastle',
    author_email='dev@aicastle.io',
    description='AI Castle Package',
    url='https://github.com/ai-castle/aicastle',
    python_requires='>=3.6',
    classifiers=[
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
    ],
    zip_safe=False,
)



import os
import shutil
from pathlib import Path
import click

@click.group()
def main():
    """AI Castle CLI"""
    pass

def copy_if_not_exists(src: Path, dst: Path):
    """
    Copy files and directories from src to dst, skipping existing files.
    """
    if src.is_dir():
        # Ensure destination directory exists
        dst.mkdir(parents=True, exist_ok=True)
        for item in src.iterdir():
            copy_if_not_exists(item, dst / item.name)
    else:
        if not dst.exists():
            shutil.copy2(src, dst)


# `chat` 명령 그룹
@main.group(invoke_without_command=True)
@click.pass_context
def chat(ctx):
    """Commands related to AI Castle Chat"""
    if ctx.invoked_subcommand is None:
        ctx.invoke(init)
        ctx.invoke(run)

# `chat init` 하위 명령
@chat.command()
def init():
    """Initialize the AI Castle Chat project."""
    source_folder = Path(__file__).parent / "package_data" / ".aicastle" / "chat"
    target_folder = Path.cwd() / ".aicastle" / "chat"  # 현재 작업 디렉터리의 .aicastle/chat

    if target_folder.exists():
        click.echo(f"Directory {target_folder} already exists! Only new files will be copied.")
    else:
        click.echo(f"Initializing AI Castle Chat at {target_folder}")

    copy_if_not_exists(source_folder, target_folder)
    click.echo(f"AI Castle Chat initialized at {target_folder}")

# `chat` 명령
@chat.command(name="run")
def run():
    """Run the AI Castle Chat application."""
    app_path = Path(__file__).parent / "chat" / "app.py"
    os.system(f"streamlit run {app_path}")

# `dev` 명령
@main.command()
def dev():
    """Copy devcontainer.json from package data."""
    source_file = Path(__file__).parent / "package_data" / ".devcontainer" / "devcontainer.json"
    target_file = Path.cwd() / ".devcontainer" / "devcontainer.json"

    if target_file.exists():
        click.echo(f"File {target_file} already exists!")
    else:
        target_file.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy2(source_file, target_file)
        click.echo(f"Copied devcontainer.json to {target_file}")